import re
import requests
import platform
import backoff
import json

from abc import ABCMeta, abstractmethod
from packaging.version import Version, parse
from pathlib import Path

from smart_webdriver_manager.cache import (
    DriverCache,
    BrowserCache,
    BrowserUserDataCache,
    DEFAULT_BASE_PATH,
)
from smart_webdriver_manager.utils import download_file

from . import logger


class SmartContextManager(metaclass=ABCMeta):
    def __init__(self, browser_name, base_path=None):
        self._base_path = base_path or DEFAULT_BASE_PATH
        self._browser_name = browser_name
        self._driver_name = self._browser_to_driver()
        self._driver_cache = DriverCache(self._driver_name, base_path)

    @property
    def driver_platform(self):
        return {"Windows": "win32", "Linux": "linux64", "Darwin": "mac64"}.get(platform.system())

    @property
    def browser_platform(self):
        return {
            "Windows": "Win_x64",
            "Linux": "Linux_x64",
            "Darwin": "Mac",
        }.get(platform.system())

    @abstractmethod
    def get_driver_release(self, version: int = 0) -> Version:
        """Translate the version to a release"""
        pass

    @abstractmethod
    def get_browser_release(self, version: int = 0) -> (Version, Version):
        """Translate the version to a release"""
        pass

    @abstractmethod
    def get_browser_user_data(self, version: int = 0) -> str:
        pass

    @abstractmethod
    def get_driver(self, release) -> str:
        pass

    @abstractmethod
    def get_browser(self, release, revision=None) -> str:
        pass

    def _browser_to_driver(self):
        if re.search(r"chrom|goog", self._browser_name, re.I):
            return "chromedriver"
        raise NotImplementedError("Other browers are not avaialble")


class SmartChromeContextManager(SmartContextManager):
    """Downloads and saves chromedriver packages
    - Manages chromedriver
    - Optionally also manages Chromium

    Chromedriver LATEST_RELEASE gives the latest version of Chromedriver, ie 96
    But if downloading Chromium, the latest is 98, not supported by Chromedriver 96

    """

    def __init__(self, base_path=None):
        super().__init__("chrome", base_path)
        self._browser_cache = BrowserCache(self._browser_name, self._base_path)
        self._browser_user_data_cache = BrowserUserDataCache(self._browser_name, self._base_path)

        self.url_driver_repo = "https://chromedriver.storage.googleapis.com"
        self.url_driver_repo_latest = f"{self.url_driver_repo}/LATEST_RELEASE"

        url_browser_repo = "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o"
        self.url_browser_zip = f"{url_browser_repo}/{self.browser_platform}%2F{{}}%2Fchrome-{{}}.zip?alt=media"

    def browser_zip(self, revision: str):
        win = lambda x: "win" if x > 591479 else "win32"  # naming changes (roughly v70)
        return {
            "Windows": win(int(revision)),
            "Linux": "linux",
            "Darwin": "mac",
        }.get(platform.system())

    @backoff.on_exception(backoff.expo, requests.exceptions.RequestException, max_time=30)
    def get_driver_release(self, version: int = 0) -> Version:
        """Find the latest driver version corresponding to the browser release"""
        logger.debug(f"Getting {self._driver_name} version for {version}")
        _version = f"_{version}" if version else ""
        url = f"{self.url_driver_repo_latest}{_version}"
        resp = requests.get(url)
        if resp.status_code == 404:
            raise ValueError(f"There is no driver for version {version}")
        elif resp.status_code != 200:
            raise ValueError(
                f"response body:\n{resp.json()}\n"
                f"request url:\n{resp.request.url}\n"
                f"response headers:\n{dict(resp.headers)}\n"
            )
        return parse(resp.text.rstrip())

    @backoff.on_exception(backoff.expo, requests.exceptions.RequestException, max_time=30)
    def get_browser_release(self, version: int = 0) -> (Version, Version):
        """Find latest corresponding chromium relese to specified/latest chromedriver
        """
        release = self.get_driver_release(version)

        revision_url = 'https://chromiumdash.appspot.com/fetch_milestones?only_branched=true'
        revisions = json.loads(requests.get(revision_url).content.decode())
        if not version:
            revisions = sorted(revisions, key=lambda x: int(x['milestone']), reverse=True)
        else:
            revisions = [d for d in revisions if int(d['milestone']) == version]
        revision = int(revisions[0]['chromium_main_branch_position'])

        while 1:
            logger.debug(f"Trying revision {revision} ... ")
            browser_zip = self.browser_zip(revision)
            url_browser_zip = self.url_browser_zip.format(revision, browser_zip)
            status_code = requests.head(url_browser_zip).status_code
            if status_code == 200:
                break
            revision -= 1

        logger.debug(f"Chromedriver version {version} supports chromium {release=} {revision=}")
        return release, parse(str(revision))

    def get_driver(self, release: str) -> Path:
        """Get driver zip for version"""
        binary_path = self._driver_cache.get(release)
        if binary_path:
            logger.debug(f"Already have latest version for release {release}")
            return binary_path

        zip_file = f"chromedriver_{self.driver_platform}.zip"
        url_zip = f"{self.url_driver_repo}/{release}/{zip_file}"
        logger.debug(f"Getting {zip_file} from {url_zip}")

        with download_file(url_zip) as f:
            binary_path = self._driver_cache.put(f, release)
            logger.debug(f"Downloaded {zip_file}")

        return binary_path

    def get_browser(self, release: str, revision: str = None) -> Path:
        """An extension of `get_supported_chromium_revision`"""
        binary_path = self._browser_cache.get(release, revision)
        if binary_path:
            logger.debug(f"Already have latest version for {release=} {revision=}")
            return binary_path

        browser_zip = self.browser_zip(revision)
        zip_file = f"{revision}-chrome-{browser_zip}.zip"
        url_zip = self.url_browser_zip.format(revision, browser_zip)
        logger.debug(f"Getting {zip_file} from {url_zip}")

        with download_file(url_zip) as f:
            binary_path = self._browser_cache.put(f, release, revision)
            logger.debug(f"Downloaded {zip_file}")

        return binary_path

    def get_browser_user_data(self, release: str, revision: str) -> Path:
        """Get browser user data dir that matches release version
        - revision is ignored (data dir is same level as major version)
        """
        data_dir_path = self._browser_user_data_cache.get(release, revision)

        return str(data_dir_path)



import os
import zipfile
import shutil
import tempfile
import requests
import backoff
import platform
import tqdm
from contextlib import contextmanager
from urllib.parse import urlparse, unquote
from pathlib import Path
from packaging.version import parse, Version

from . import logger


class LinuxZipFileWithPermissions(zipfile.ZipFile):
    """Class for extract files in linux with right permissions
    https://stackoverflow.com/a/54748564
    """

    def _extract_member(self, member, targetpath=None, pwd=None):
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)
        if targetpath is None:
            targetpath = os.getcwd()
        target = super()._extract_member(member, targetpath, pwd)
        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(target, attr)
        return target


def unpack_zip(zip_path):
    """Unzip zip to same diretory"""
    zip_class = zipfile.ZipFile if platform.system() == "Windows" else LinuxZipFileWithPermissions
    archive = zip_class(zip_path)
    try:
        archive.extractall(Path(zip_path).parent)
    except Exception as e:
        if e.args[0] not in [26, 13] and e.args[1] not in ["Text file busy", "Permission denied"]:
            raise e
    return archive.namelist()


@contextmanager
def download_file(url) -> Path:
    """Better download"""
    name = Path(urlparse(unquote(url)).path).name
    with mktempdir() as tmpdir:

        @backoff.on_exception(backoff.expo, requests.exceptions.RequestException, max_time=30)
        def get():
            with requests.get(url, stream=True) as r:
                save_path = tmpdir.joinpath(name)
                total = int(r.headers.get('content-length', 0))
                chunk = 16 * 1024 * 1024
                with open(save_path, "wb") as f, tqdm.tqdm(total=total, desc=name, unit='B', unit_scale=True) as p:
                    while buf := r.raw.read(chunk):
                        f.write(buf)
                        p.update(len(buf))
                return save_path

        yield get()


@contextmanager
def mktempdir() -> Path:
    """Having errors removing temp directories in Widnows..."""
    try:
        tmp = tempfile.mkdtemp()
        yield Path(tmp)
    finally:

        @backoff.on_exception(backoff.expo, shutil.Error, max_time=10)
        def remove():
            shutil.rmtree(tmp, ignore_errors=False)
            logger.debug(f"Removed {tmp}")

        remove()



from abc import ABCMeta, abstractmethod
from functools import cache

from smart_webdriver_manager.context import SmartChromeContextManager

from . import logger


class DriverManager(metaclass=ABCMeta):
    def __init__(self, version, base_path):
        self._base_path = base_path
        self._version = version
        logger.info("Running driver manager")

    @abstractmethod
    def get_driver(self) -> str:
        pass

    @abstractmethod
    def get_browser(self) -> str:
        pass

    @abstractmethod
    def get_browser_user_data(self) -> str:
        pass

    # @abstractmethod
    # def remove():
    # pass


class ChromeDriverManager(DriverManager):

    __called_driver__ = False
    __called_browser__ = False

    def __init__(self, version: int = 0, base_path=None):
        super().__init__(version, base_path)
        self._cx = SmartChromeContextManager(self._base_path)

    @cache
    def get_driver(self):
        """Smart lookup for current driver version
        - chromedriver version will always be <= latest chromium browser
        """
        driver_release = self._cx.get_driver_release(self._version)
        driver_path = self._cx.get_driver(str(driver_release))
        self.__called_driver__ = True
        if not self.__called_browser__:
            self.get_browser()
        return str(driver_path)

    @cache
    def _get_browser_helper(self):
        if not self.__called_driver__:
            self.get_driver()
        return self._cx.get_browser_release(self._version)

    @cache
    def get_browser(self):
        self.__called_browser__ = True
        browser_release, browser_revision = self._get_browser_helper()
        browser_path = self._cx.get_browser(str(browser_release), str(browser_revision))
        return str(browser_path)

    @cache
    def get_browser_user_data(self):
        if not self.__called_browser__:
            self.get_browser()
        browser_release, browser_revision = self._get_browser_helper()
        user_data_path = self._cx.get_browser_user_data(str(browser_release), str(browser_revision))
        return str(user_data_path)

    # def _forfun(self):
        # # test and remove me
        # return [
            # f'driver stats:  {self.get_driver.cache_info()}',
            # f'browser stats: {self.get_browser.cache_info()}',
            # f'datadir stats: {self.get_user_data.cache_info()}',
            # ]

    # def remove(self):
        # """Needs to remove all the instances from cache (force kill if running)
        # """
        # self.get_driver.cache_clear()
        # self.get_browser.cache_clear()
        # self.get_user_data.cache_clear()
        # self._get_browser_helper.cache_clear()
        # self.__called_driver__ = False
        # self.__called_browser__ = False
        # # now remove things



__version__ = "0.5.0"


import logging

logger = logging.getLogger(__name__)


# Avaialble managers
from smart_webdriver_manager.driver import ChromeDriverManager


__all__ = [
    "ChromeDriverManager",
]



import datetime
import json
import re
import platform
import glob

from abc import ABCMeta, abstractmethod
from pathlib import Path
from smart_webdriver_manager.utils import unpack_zip

from . import logger


DEFAULT_BASE_PATH = {
    "Windows": Path("~/appdata/roaming/swm").expanduser(),
    "Linux": Path("~/.local/share/swm").expanduser(),
    "Drawin": Path("~/Library/Application Support/swm").expanduser(),
}.get(platform.system(), Path("~/.swm").expanduser())


class SmartCache(metaclass=ABCMeta):
    """Shared Cache parent, controls cache behavior"""

    def __init__(self, cache_name, base_path=None):
        self._base_path = Path(base_path or DEFAULT_BASE_PATH).expanduser()
        self._cache_json_path = self._base_path.joinpath(f"{cache_name}.json")
        self._cache_base_path = self._base_path.joinpath(cache_name)

    @abstractmethod
    def get(self, typ, release, revision=None) -> Path:
        metadata = self._read_metadata()
        key = f"{typ}_{release}{'_' if revision else ''}{revision or ''}"
        if key not in metadata:
            logger.info(f"There is no {key}, {release}, {revision=} in cache")
            return
        driver_info = metadata[key]
        path = driver_info["binary_path"]
        logger.info(f"{key} found in cache at path {path}")
        return Path(path)

    @abstractmethod
    def put(self, f, typ, release, revision=None) -> Path:
        path = Path(self._cache_base_path, typ, release, revision or "")
        path.mkdir(parents=True, mode=0o755, exist_ok=True)

        f = Path(f)
        zip_path = f.replace(path.joinpath(f.name))
        logger.debug("Unzipping...")
        files = unpack_zip(zip_path)

        binary = self._match_binary(files, typ)
        binary_path = Path(path, binary)
        self._write_metadata(binary_path, typ, release, revision)
        logger.info(f"{typ} has been saved in cache at path {path}")
        return binary_path

    def _match_binary(self, files: list, typ: str) -> Path:
        logger.debug(f"Matching {typ} in candidate files")
        if len(files) == 1:
            return files[0]
        for f in files:
            name = Path(f).name
            # FIXME: Mac will not return the correct app
            re_match = re.compile(r"(ium)?(.(exe|app))?$")
            if f'{re_match.sub("", name).lower()}' in f"{typ}":
                return Path(f)
        raise Exception(f"Can't get binary for {typ} among {files}")

    def _write_metadata(self, binary_path, typ, release, revision):
        metadata = self._read_metadata()
        key = f"{typ}_{release}{'_' if revision else ''}{revision or ''}"
        data = {
            key: {
                "timestamp": datetime.date.today().strftime("%m/%d/%Y"),
                "binary_path": str(binary_path),
            }
        }
        metadata.update(data)
        with open(self._cache_json_path, "w+") as outfile:
            json.dump(metadata, outfile, indent=4)

    def _read_metadata(self):
        if Path(self._cache_json_path).exists():
            with open(self._cache_json_path, "r") as outfile:
                return json.load(outfile)
        return {}


class DriverCache(SmartCache):
    """Driver Cache"""

    def __init__(self, driver_name, base_path=None):
        super().__init__("drivers", base_path)
        self._driver_name = driver_name

    def get(self, release):
        return super().get(self._driver_name, release)

    def put(self, f, release):
        return super().put(f, self._driver_name, release)


class BrowserCache(SmartCache):
    """Browser Cache"""

    def __init__(self, browser_name, base_path=None):
        super().__init__("browsers", base_path)
        self._browser_name = browser_name

    def get(self, release, revision=None):
        return super().get(self._browser_name, release, revision)

    def put(self, f, release, revision=None):
        return super().put(f, self._browser_name, release, revision)


class BrowserUserDataCache:
    """Browser User Data Cache"""

    def __init__(self, browser_name, base_path=None):
        self._browser_cache = BrowserCache(browser_name, base_path)

    def get(self, release, revision=None):
        browser_path = self._browser_cache.get(release, revision)
        if not browser_path:
            raise AssertionError("get_browser() not yet called")
        user_data_path = Path(*browser_path.parts[: browser_path.parts.index(release) + 1])
        user_data_path = user_data_path.joinpath("UserData")
        user_data_path.mkdir(mode=0o755, exist_ok=True)
        logger.info(f"Got user data {user_data_path} for {self._browser_cache._browser_name}")
        return user_data_path



from abc import ABCMeta, abstractmethod
from smart_webdriver_manager.context import SmartChromeContextManager

from . import logger


class BrowserManager(metaclass=ABCMeta):
    @abstractmethod
    def install(self, version: int = 0, **kwargs):
        """One of the `smart` elements. Uses `version` to determine
        which driver to install
        """
        pass


class ChromeBrowserManager(BrowserManager):
    def __init__(self, base_path=None):
        self._cx = SmartChromeContextManager(base_path)
        logger.info("Running chrome browser manager")

    def install(self, version: int = 0):
        """Smart lookup for current browser version"""
        logger.debug(f"Fetching chromium browser")
        revision = self._cx.find_supported_chromium_revision(version)
        return self._cx.fetch_browser(revision)
